from .interface import ImageTestInterface
from unittest import TestCase


class CIFAR10_Test(TestCase,ImageTestInterface):

    dataset_name = 'CIFAR10'

    @classmethod
    def setUpClass(cls):
        import random
        from torchvision import transforms as T
        dataset = cls.dataset()
        cls._error_locs = list( set( [ random.randint(0,len(dataset)) for _ in range( random.randint(12,25) ) ] ) )
        cls._t = T.Compose([ 
                T.ToTensor(),
                # T.Resize( (32,32) )
         ])
    
    @classmethod
    def _get_loss(cls):
        from torch import nn
        return nn.CrossEntropyLoss()

    def _get_model(self):
        from torch.nn import functional as F
        from torch import nn
        class Net(nn.Module):
            def __init__(self):
                super(self.__class__, self).__init__()
                self.conv1 = nn.Conv2d(3, 6, 5)
                self.conv2 = nn.Conv2d(6, 16, 5)
                self.fc1   = nn.Linear(16*5*5, 120)
                self.fc2   = nn.Linear(120, 84)
                self.fc3   = nn.Linear(84, 10)

            def forward(self, x):
                out = F.relu(self.conv1(x))
                out = F.max_pool2d(out, 2)
                out = F.relu(self.conv2(out))
                out = F.max_pool2d(out, 2)
                out = out.view(out.size(0), -1)
                out = F.relu(self.fc1(out))
                out = F.relu(self.fc2(out))
                out = self.fc3(out)
                return out
        return Net()

    def test_outlier_detection(self):
        from torch.utils.data import DataLoader
        import torch
        from tqdm import tqdm
        dataset = self.dataset()
        mw = self._init_outlier_detection()
        data_set = DataLoader( dataset,batch_size=16,shuffle=False )
        model = self.model()
        optim = self.optimizer()
        loss = self.loss()
        for idx,(x,y) in enumerate(tqdm(data_set,total=int(len(dataset)))):
            if idx in self._error_locs:
                x += torch.randn( x.size() )
                y += torch.ones( y.size() ).long()
                y = y % 10
            optim.zero_grad()
            y_hat = model(x)
            mw( y_hat,y,loss ).backward()
            optim.step()
        self.assertTrue(
            len(list(mw.montitors.values())[0].oods ) > 0,
            'No Out Of Distribution Samples weere found'
        )
        self.assertGreaterEqual(
            round((len(set(list(mw.montitors.values())[0].oods  ).intersection( self._error_locs ))/len(self._error_locs) ) * 100),
            12.5,
            'Did not find enough Out Of Distribution samples'
        )
        self.assertLessEqual(
            round((len(set(list(mw.montitors.values())[0].oods  ).intersection( self._error_locs ))/len(self._error_locs) ) * 100),
            25,
            f'Catching too many Out Of Distribution samples, logged {round((len(set(list(mw.montitors.values())[0].oods  ).intersection( self._error_locs ))/len(self._error_locs) ) * 100)}'
        )


    

    




import torch
from .interface import BaseExplainerInterface

class TabularExplainer(BaseExplainerInterface):
    """
    

    :param BaseExplainerInterface: [description]
    :type BaseExplainerInterface: [type]
    """    

    def __init__(self):
        super(self.__class__,self).__init__()

        






def make_classifier():
    from tests.manual import constants
    from torch import nn
    return nn.Sequential(
        nn.Linear(constants.INPUT_SIZE, constants.INPUT_SIZE),
        nn.ReLU(),
        nn.Linear(constants.INPUT_SIZE, constants.N_CLASSES),
    )

def make_data():
    from tests.manual import constants
    import torch
    return torch.randn((8,constants.INPUT_SIZE)),torch.randn((8,constants.N_CLASSES))



from typing import Union,Optional
from metrician.configs.cfg import DefaultCFG
from torch import Tensor
from metrician.writers.interface import BaseWriterInterface
from metrician.signals import SignalDataKeys
MetricWriter = None
class MetricWriter(BaseWriterInterface):
    """
        Metric Writer is a lightweight class that will automatically log metrics during a training loop.
    """
        
    def __init__(   self,
                    cfg:Union[dict,DefaultCFG] = None
                ): 
        """

        Args:
            cfg (Union[dict,DefaultCFG], optional): [description]. Defaults to None.
        """        
        super(self.__class__,self).__init__()
        
        import sklearn.metrics as metrics
        from metrician.configs.interface import DefaultConfigInterface

        # check for user provided config, 
        # if not present fallback to default
        self.cfg = DefaultCFG( cfg if isinstance( cfg, dict ) else None ) if not issubclass(cfg.__class__,DefaultConfigInterface) else cfg
        
        self.metrics = metrics
        self.custom_functions = {}
        self.signals = {}
        self.montiors = {}
        self.index = 0

    def forward(   self,
                    _output : Tensor, 
                    _labels: Tensor, 
                    loss_fn: Optional[callable] = None,
                    index : Optional[int] = -1
                ) -> Union[ 'loss_fn',MetricWriter ]:

        """[summary]

        Returns:
            [type]: [description]
        """        
        self.index += 1
        output,labels = self.cfg.sanitize_inputs( _output, _labels )
        for metric in self.cfg.metrics:
            if hasattr( self.metrics,metric.name ): self.writer.add_scalar(
                self.cfg.main_tag+'_'+metric.name,
                getattr(self.metrics,metric.name)( output,labels ) if metric.name != 'f1_score' else getattr(self.metrics,metric.name)( output,labels, average='micro'),
            index if index > -1 else self.index
        )
        for name,func in self.custom_functions.items():
            self.writer.add_scalar(
                                    self.cfg.main_tag+'_'+name,
                                    func( output,labels ),
                                    index if index > -1 else self.index
            )
        if callable(loss_fn):
            loss = loss_fn( _output,_labels )
            self.writer.add_scalar(self.cfg.main_tag+'_loss',loss.item(),index if index > -1 else self.index)
            if len(self.montitors) > 0:
                l = loss.item()
                pckt = {
                    SignalDataKeys.LOSS:l,
                    SignalDataKeys.PREDICTIONS:_output,
                    SignalDataKeys.LABELS:_labels
                }
                for k,v in self.montitors.items():v( **pckt )
            return loss

        return self



from torch.utils import data
from typing import Union
class Dataset:

    def __init__(
                    self,
                    dataset : data.Dataset,
                    callbacks : list = []
                ):
        self.dataset = dataset
        self.callbacks = callbacks

    def __getitem__(self,index : Union[slice,int]):
        return self.dataset[index]


from .interface import ImageTestInterface
from unittest import TestCase


class MNIST_Test(TestCase,ImageTestInterface):
    @classmethod
    def setUpClass(cls):
        import random
        from torchvision import transforms as T
        dataset = cls.dataset()
        cls._error_locs = list( set( [ random.randint(0,len(dataset)) for _ in range(25) ] ) )
        cls._t = T.Compose([ 
                T.ToTensor(),
                T.Resize( (28,28) )
         ])

    def _get_model(self):
        from torch.nn import functional as F
        from torch import nn
        class Net(nn.Module):
            def __init__(self):
                super(Net, self).__init__()
                self.conv1 = nn.Conv2d(1, 32, 3, 1)
                self.conv2 = nn.Conv2d(32, 64, 3, 1)
                self.dropout1 = nn.Dropout(0.25)
                self.dropout2 = nn.Dropout(0.5)
                self.fc1 = nn.Linear(9216, 128)
                self.fc2 = nn.Linear(128, 10)

            def forward(self, x):
                x = self.conv1(x)
                x = F.relu(x)
                x = self.conv2(x)
                x = F.relu(x)
                x = F.max_pool2d(x, 2)
                x = self.dropout1(x)
                x = torch.flatten(x, 1)
                x = self.fc1(x)
                x = F.relu(x)
                x = self.dropout2(x)
                x = self.fc2(x)
                output = F.log_softmax(x, dim=1)
                return output
        return Net()

    def test_outlier_detection(self):
        from torch.utils.data import DataLoader
        from tqdm import tqdm
        dataset = self.dataset()
        mw = self._init_outlier_detection()
        data_set = DataLoader( dataset,batch_size=4 )
        model = self.model()
        optim = self.optimizer()
        loss = self.loss()
        for idx,(x,y) in enumerate(tqdm(data_set,total=len(dataset))):
            
            if idx in self._error_locs:x *= torch.randn( x.size() )
            optim.zero_grad()
            y_hat = model(x,y)
            mw( y_hat,y,loss ).backward()
            optim.step()



    

    




from .interface import BaseMonitorInterface
from typing import List

from torch.utils.data import Dataset
from typing import Union,Iterable
from metrician.signals.outliers import MomentOutlierSignal as MOS
from metrician.signals.interface import BaseSignal
from metrician.recorders.interface import BaseRecorderInterface
from metrician.signals import SignalDataKeys
from metrician.explainations.interface import BaseExplainerInterface
from torch import Tensor
import metrician
class OutOfDistributionMonitor(BaseMonitorInterface):
    """
        Default monitor for out of distribution losses.

        Basically this sets up a monitor that calculates whether some output is an outlier and records the input so you can evaluate it later on

        :param BaseMonitorInterface: [description]
        :type BaseMonitorInterface: [type]
    """        

        def __init__(   
                    self,
                    dataset : Union[Dataset,Iterable],
                    data_type : Union[str,metrician.DatasetType] = None,
                    signals  : List[BaseSignal] = [MOS()],
                    recorders : List[BaseRecorderInterface] = [lambda x: x],
                    explainer  : BaseExplainerInterface = None,
                    slug : str = 'Out Of Distribution Sample'
                ):
        """

        :param dataset: [description]
        :type dataset: Union[Dataset,Iterable]
        :param data_type: [description], defaults to None
        :type data_type: Union[str,metrician.DatasetType], optional
        :param signal: [description], defaults to MOS()
        :type signal: BaseSignal, optional
        :param recorder: [description], defaults to None  
        :type recorder: BaseRecorderInterface, optional
        :param explainer: [description], defaults to None
        :type explainer: BaseExplainerInterface, optional
        :param slug: [description], defaults to 'Out Of Distribution Sample'
        :type slug: str, optional
        """                       
        super( self.__class__,self ).__init__()     
        assert data_type is None and dataset.__class__.__name__ in metrician.DATASET_INPUT_TYPES, 'If your dataset is not a torch dataset you have to specify the input data type [image,audio,text,tabular] '
        self.dataset_length = len(dataset)
        self.dtype = data_type if data_type is not None else metrician.DATASET_INPUT_TYPES[dataset.__class__.__name__]
        self.signals = signals
        self._index = 0
        self.dataset = dataset
        self.slug = slug
        self.writer = self._write_fn()
        self.recorders = recorders
        self.explainer = explainer

    def _check_signal(self,signal : BaseSignal,*args,**kwargs) -> None:
        """[summary]

        :param signal: [description]
        :type signal: BaseSignal
        """      

        ood = signal( *args,**kwargs )
        if ood:
            for recorder in self.recorders:
                self.writer(
                    self.slug+'_'+signal.__class__.__name__,
                    # take the first index since it's the input
                    recorder( self.dataset[ self._index % self.dataset_length ][0] ), 
                    self._index
                )
            if self.explainer is not None: 
                kwargs[SignalDataKeys.INPUT] = self.dataset[ self._index % self.dataset_length ][0]
                self.explainer( kwargs )

            self.captured_signals.append( self._index % self.dataset_length )


    def forward(self,*args, **kwargs) -> None:
        """


        

        :param loss: output from loss function
        :type loss: Tensor
        :param y: data labels/target
        :type y: Tensor
        :param yhat: model's prediction 
        :type yhat: Tensor
        """                 
        for signal in self.signals:self._check_signal(signal,*args, **kwargs)
        
        self._index += 1


    




# from unittest import TestCase
from abc import ABC

class BaseTestInterface(object):
    
    _loss       = None
    _model      = None
    _index      = 0
    _dataset    = None
    _optimizer  = None
    _task_type  = None
    _error_locs = None


    @classmethod
    def tearDownClass(cls):
        cls._loss       = None
        cls._model      = None
        cls._index      = 0
        cls._dataset    = None
        cls._optimizer  = None
        cls._task_type  = None
        cls._error_locs = None

    @classmethod
    def _preprocess(cls,data): raise NotImplementedError('You must create a custom dataset builder')

    @classmethod
    def _postprocess(cls,data): return data#raise NotImplementedError('You must create a custom dataset builder')

    @classmethod
    def _get_dataset(cls):raise NotImplementedError('You must create a custom dataset builder')

    @classmethod
    def _get_model(cls):raise NotImplementedError('You must create a custom model')

    @classmethod
    def _get_loss(cls):raise NotImplementedError('You must create a custom loss getter')

    @classmethod
    def _get_optimizer(cls):#raise NotImplementedError('You must create a custom optimizer getter')
        from torch.optim import SGD
        model = cls.model()
        return SGD( model.parameters(),lr=.01 )
    
    # @property
    @classmethod
    def dataset(cls): 
        if cls._dataset is None:cls._dataset=cls._get_dataset()
        return cls._dataset

    # @property
    @classmethod
    def loss(cls): 
        if cls._loss is None:cls._loss=cls._get_loss()
        return cls._loss

    # @property
    @classmethod
    def optimizer(cls): 
        if cls._optimizer is None:cls._optimizer=cls._get_optimizer()
        return cls._optimizer    

    # @property
    @classmethod
    def model(cls): 
        if cls._model is None:cls._model=cls._get_model(cls)
        return cls._model

    @classmethod
    def _init_outlier_detection(cls):
        from metrician import MetricWriter,SimpleClf,OutOfDistributionMonitor
        dataset = cls.dataset()
        ood = OutOfDistributionMonitor(dataset)
        ood.testing = True
        return MetricWriter( SimpleClf() ).register_monitor(
            ood
        )

    @classmethod
    def _check_outliers(cls):pass

            
        

    






class AutoDict(dict):
    """
        Allows for JS Object type interaction
        Recusrively sets attributes as a dictionary
    """
    def __init__(self,**kwargs):
        for k,v in kwargs.items():setattr(self,k,v if not isinstance(v, (dict) ) else AutoDict(**v) )
    

def DefaultCFG(AutoDict) -> AutoDict:

    def __init__(self,cfg : dict = {}) -> 'DefaultCFG':
        super(self,DefaultCFG).__init__( **cfg )
        self.validate_config(self)

    @staticmethod
    def validate_config(cls : DefaultCFG ) -> bool:
        """
            Iterate over config and validate data structure
        """
        pass



import unittest

class BasicTest(unittest.TestCase):

    def test_writer(self):
        from tests.manual.utils import make_classifier,make_data
        from src.configs.defaults import SimpleClf
        from src.writers.writer import MetricWriter
        from torch import nn
        loss = nn.L1Loss() # it doesn't really matter all that much for right now, though later I'll test them all
        clf = make_classifier()
        mw = MetricWriter( SimpleClf )
        self.assertTrue( mw.cfg is not None,mw.cfg )
        for _ in range(20):
            x,y = make_data()
            y_hat = clf(x)
            mw( y_hat, y, loss )
        mw.close()
        return True


if __name__ == "__main__":
    unittest.main()




from .interface import BaseSignal
from metrician.signals import SignalDataKeys
from torch import Tensor
import math
class LossSignal(BaseSignal):
    """
        The Loss Signal is basically the Moment Outlier, but with a Pareto distribution to account for the decreasing loss value

        :param BaseSignal: [description]
        :type BaseSignal: [type]
    """    

    keys = [ SignalDataKeys.LOSS ]

    

    def __init__(self,alpha : float = 4.25):
        """[summary]

        :param alpha: [description], defaults to 2.0
        :type alpha: float, optional
        """        
        super( self.__class__,self ).__init__()          
        self._index = self._eta = self._rho = self._tau = self._phi = 0.0
        self._min   = self._max = float('nan')
        self.n_samples_needed = 10000000
        self._alpha = alpha

    def forward(self, *args,**kwargs) -> bool:
        '''
            :param x: torch.Tensor
            :returns bool: if the input tensor is an outlier
        '''       
        self._check_input( kwargs )
        x = kwargs[self.keys[0]]
        if self._index == 0.0:
            self._min = x
            self._max = x
        else:
            # TODO: abstract for multidimensional input
            self._min = min(self._min, x)
            self._max = max(self._max, x)
        
        
        delta = ( ( x*self._alpha )/(self._alpha) ) - self._eta
        delta_n = delta / (self._index + 1)
        delta_n2 = delta_n * delta_n
        term = ((x**2)*self._alpha)/( ((self._alpha-1)**2)*(self._alpha - 2) )

        self._eta += delta_n
        self._phi += (
            term * delta_n2 * (self._index ** 2 - 3 * self._index + 3)
            + 6 * delta_n2 * self._rho
            - 4 * delta_n * self._tau
        )
        self._tau += (
            term * delta_n * (self._index - 2) - 3 * delta_n * self._rho
        )
        self._rho += term
        self._index += 1
        self.n_samples_needed = self._calc_sample_size()
        outlier = self._is_outlier(x) if self._index > self.n_samples_needed else False
        return outlier

    def _is_outlier(self, x : Tensor) -> bool:
        '''
            weight the amount of standard deviation to use based on kurtosis of the data

            :param x: Pytorch Tensor
            :returns bool: Whether the inputed tensor is an outlier or not
        '''
        m = self.mean
        std = math.sqrt( self.var )
        k = self.kurtosis
        shift = std*( 1+k )
        #               Above the threshold           |         Below the threshold
        return not bool( ( m + shift ) >= x >= ( m - shift)  )

    def _calc_sample_size(self) -> float:
        '''
            Calculate the necessary sample size for a distribution using the z-score
        '''
        import math
        # 90% confidence interval: 1.645^2 = 2.7225
        std = math.sqrt( self._rho / (self._index  ) if self._index > 1 else self._index )
        return int( (2.7225 * std * (1-std))/ .01)+1

    @property
    def mean(self) -> Tensor: return self._eta

    @property
    def var(self) -> Tensor: return self._rho / (self._index )

    @property
    def skewness(self) -> Tensor: return (self._index ** 0.5) * self._tau / pow(self._rho, 1.5)

    @property
    def kurtosis(self) -> Tensor: return (self._index * self._phi / (self._rho * self._rho) - 3.0)/(self._index+1)


    


from enum import Enum
class DatasetType(Enum):
    audio       = 'audio'
    image       = 'image'
    text        = 'text'
    tabular     = 'distribution'

IMAGE_DATASETS = '''CelebA
CIFAR10
CIFAR100
Cityscapes
COCO
DatasetFolder
EMNIST
FakeData
Fashion-MNIST
Flickr
HMDB51
ImageFolder
ImageNet
Kinetics-400
KMNIST
LSUN
MNIST
Omniglot
PhotoTour
Places365
QMNIST
SBD
SBU
STL10
SVHN
UCF101
USPS
VOC'''.split('\n')
AUDIO_DATASETS = '''CMUARCTIC
COMMONVOICE
GTZAN
LIBRISPEECH
LIBRITTS
LJSPEECH
SPEECHCOMMANDS
TEDLIUM
VCTK
VCTK_092
YESNO'''.split('\n')
TEXT_DATASETS = '''WikiText-2
WikiText103
PennTreebank
SST
IMDb
TextClassificationDataset
AG_NEWS
SogouNews
DBpedia
YelpReviewPolarity
YelpReviewFull
YahooAnswers
AmazonReviewPolarity
AmazonReviewFull
TREC
SNLI
MultiNLI
Multi30k
IWSLT
WMT14
UDPOS
CoNLL2000Chunking
BABI20
EnWik9'''.split('\n')
DATASET_INPUT_TYPES = {

}

DATASET_INPUT_TYPES.update( {ds:DatasetType.audio for ds in AUDIO_DATASETS } )
DATASET_INPUT_TYPES.update( {ds:DatasetType.image for ds in IMAGE_DATASETS } )
DATASET_INPUT_TYPES.update( {ds:DatasetType.text for ds in TEXT_DATASETS } )




# from .interface import BaseSignal
# from torch import Tensor
# import math
# class MomentOutlierSignal(BaseSignal):
#     '''
#         Real time outlier detection for metrics

#         This class records moments in order to estimate the likelihood of an outlier when data is passed to it. 
#         If data may be an outlier, such as something that causes a really high error, this will record the index so you can view it later
#     '''

#     def __init__(self):
#         self._index = self._eta = self._rho = self._tau = self._phi = 0.0
#         self._min   = self._max = float('nan')

#     def __call__(self, x : Tensor) -> int:
#         if self._index == 0.0:
#             self._min = x
#             self._max = x
#         else:
#             # TODO: abstract for multidimensional input
#             self._min = min(self._min, x)
#             self._max = max(self._max, x)
        
#         outlier = self._is_outlier(x) if self._index > 5 else False
#         if not outlier: # don't include the outliers
#             delta = x - self._eta
#             delta_n = delta / (self._index + 1)
#             delta_n2 = delta_n * delta_n
#             term = delta * delta_n * self._index

#             self._eta += delta_n
#             self._phi += (
#                 term * delta_n2 * (self._index ** 2 - 3 * self._index + 3)
#                 + 6 * delta_n2 * self._rho
#                 - 4 * delta_n * self._tau
#             )
#             self._tau += (
#                 term * delta_n * (self._index - 2) - 3 * delta_n * self._rho
#             )
#             self._rho += term
#         self._index += 1
#         return 1 if outlier else 0

#     def _is_outlier(self, x : Tensor) -> bool:
#         '''
#             weight the amount of standard deviation to use based on kurtosis of the data
#         '''
#         m = self.mean
#         std = math.sqrt( self.var )
#         k = self.kurtosis
#         #            Above the threshold           |         Below the threshold
#         return ( x > ( m + (std*(3.5 - k ) ) )  ) or ( x < ( m - (std*(3.5 - k ) ) )  )


#     @property
#     def mean(self) -> Tensor: return self._eta

#     @property
#     def var(self) -> Tensor: return self._rho / (self._index - 1)

#     @property
#     def skewness(self) -> Tensor: return (self._index ** 0.5) * self._tau / pow(self._rho, 1.5)

#     @property
#     def kurtosis(self) -> Tensor: return (self._index * self._phi / (self._rho * self._rho) - 3.0)/2.


    



from .interface import BaseSignal
from torch import Tensor
import math
class MomentOutlierSignal(BaseSignal):
    """
        Real time outlier detection for metrics

        This class records moments in order to estimate the likelihood of an outlier when data is passed to it. 
        If data may be an outlier, such as something that causes a really high error, this will record the index so you can view it later.

        NOTE:   This Signal assumes that the underlying generating function of the data is stationary. 
                If you'd like measure the outliers for your loss function, please use the "Loss" Signal.
                The Loss Signal captures a Pareto distribution - the expected value is monotonically decreasing.


        :param BaseSignal: [description]
        :type BaseSignal: [type]
    """    

    

    def __init__(self):
        super( self.__class__,self ).__init__()
        """[summary]
        """        
        self._index = self._eta = self._rho = self._tau = self._phi = 0.0
        self._min   = self._max = float('nan')
        self.n_samples_needed = 10000000

    def forward(self, x : Tensor) -> bool:
        '''
            :param x: torch.Tensor
            :returns bool: if the input tensor is an outlier
        '''       
        if self._index == 0.0:
            self._min = x
            self._max = x
        else:
            # TODO: abstract for multidimensional input
            self._min = min(self._min, x)
            self._max = max(self._max, x)
        
        
        # if not outlier: # don't include the outliers
        delta = x - self._eta
        delta_n = delta / (self._index + 1)
        delta_n2 = delta_n * delta_n
        term = delta * delta_n * self._index

        self._eta += delta_n
        self._phi += (
            term * delta_n2 * (self._index ** 2 - 3 * self._index + 3)
            + 6 * delta_n2 * self._rho
            - 4 * delta_n * self._tau
        )
        self._tau += (
            term * delta_n * (self._index - 2) - 3 * delta_n * self._rho
        )
        self._rho += term
        self._index += 1
        self.n_samples_needed = self._calc_sample_size()
        outlier = self._is_outlier(x) if self._index > self.n_samples_needed else False
        return outlier

    def _is_outlier(self, x : Tensor) -> bool:
        '''
            weight the amount of standard deviation to use based on kurtosis of the data

            :param x: Pytorch Tensor
            :returns bool: Whether the inputed tensor is an outlier or not
        '''
        m = self.mean
        std = math.sqrt( self.var )
        k = self.kurtosis
        shift = std*( 1+k )
        #               Above the threshold           |         Below the threshold
        return not bool( ( m + shift ) >= x >= ( m - shift)  )

    def _calc_sample_size(self) -> float:
        '''
            Calculate the neccessary sample size for a distribution using the z-score
        '''
        import math
        # 90% confidence interval: 1.645^2 = 2.7225
        std = math.sqrt( self._rho / (self._index -1 ) if self._index > 1 else self._index )
        return int( (2.7225 * std * (1-std))/ .01)+1

    @property
    def mean(self) -> Tensor: return self._eta

    @property
    def var(self) -> Tensor: return self._rho / (self._index - 1)

    @property
    def skewness(self) -> Tensor: return (self._index ** 0.5) * self._tau / pow(self._rho, 1.5)

    @property
    def kurtosis(self) -> Tensor: return (self._index * self._phi / (self._rho * self._rho) - 3.0)


    



from metrician.configs.interface import DefaultConfigInterface,ConfigType
from metrician.configs.cfg import AutoDict

class SimpleClf(DefaultConfigInterface):
    """
        Default metrics for classification are F1, Accuracy and AUC-ROC
    """

    main_tag        = 'Classifier'
    metrics         = [ AutoDict(name=fn) for fn in 'accuracy_score,f1_score,roc'.split(',') ]
    config_type     = ConfigType.classification

class SimpleRgr(DefaultConfigInterface):
    """
        Default metrics for regression are Explained Variance, R2, Mean Absolute Error and Median Absolute Error
    """

    main_tag        = 'Regressor'
    metrics         = [ AutoDict(name=fn) for fn in 'explained_variance_score,r2_score,median_absolute_error,mean_absolute_error'.split(',') ]
    config_type     = ConfigType.regression



import unittest
from unittest import TestCase
import torch
from .outliers import MomentOutlierSignal
from .loss import LossSignal
from pylab import *
import math
class LossTest(TestCase):

    def test_scalar_mean(self):
        
        # create pareto like distribution
        r = torch.randn( 100 )
        x = torch.arange( 100 ).float()+1
        x = (2.71/(torch.log( x )+.165))
        x += r*.75 # throw in some noise
        signal      =  LossSignal() # MomentOutlierSignal()#
        outliers    = [ ]
        means       = []
        var         = [[],[]]
        ks = []
        for i,x_i in enumerate(r):
            is_outlier = signal(x_i) 
            if is_outlier: outliers.append(i)
            means.append( float( signal.mean ) )
            var[0].append( float( signal.mean + math.sqrt(signal.var) ) )
            var[1].append( float( signal.mean - math.sqrt(signal.var) ) )
            ks.append( signal.kurtosis )
        self.assertLessEqual(
            len(outliers)/len(x),
            .32,
            'Too many Loss outliers found'
        )
        
        


class MomentOutlierTest(TestCase):
    
    def setUp(self):
        
        self.eps = 1e-2 # threshold for differences in data evaluation
    
    def test_scalar_mean(self):
        
        Signal = MomentOutlierSignal()
        data = torch.randn( 100 )
        mean = data.mean()
        for d in data:Signal( d )
        self.assertTrue(
            torch.abs(Signal.mean - mean) < self.eps,
            'Outlier Scalar Mean not within threshold value'
        )
    def test_scalar_var(self):
        
        Signal = MomentOutlierSignal()
        data = torch.randn( 100 )
        var = data.var()
        for d in data:Signal( d )
        self.assertTrue(
            torch.abs(Signal.var - var) < self.eps,
            'Outlier Scalar Variance not within threshold value'
        )
    def test_scalar_skewness(self):
        from scipy.stats import norm, skew
        Signal = MomentOutlierSignal()
        data = torch.randn( 100 )
        s = skew(data)
        for d in data:Signal( d )
        self.assertTrue(
            torch.abs(Signal.skewness - s) < self.eps*10,
            'Outlier Scalar Skewness not within threshold value'
        )
    def test_scalar_kurtosis(self):
        from scipy.stats import norm, kurtosis
        Signal = MomentOutlierSignal()
        data = torch.randn( 100 )
        k = kurtosis(data)
        for d in data:Signal( d )
        self.assertTrue(
            torch.abs(Signal.kurtosis - k) < self.eps*10,
            'Outlier Scalar Kurtosis not within threshold value'
        )

if __name__ == '__main__':
    unittest.main()  



import inspect
from typing import Union
from tensorboardX import GlobalSummaryWriter,SummaryWriter
BaseInterface=None
class BaseInterface(object):

    callbacks       = {}
    callback_kwargs = {}

    dtype           = None

    # @classmethod
    def _write_fn(cls) -> callable:
        from metrician import DatasetType
        writer = GlobalSummaryWriter.getSummaryWriter()
        return {
            # 'audio':writer.add_audio,
            DatasetType.text:writer.add_text,
            # 'tabular':writer.add_histogram,
            DatasetType.image:writer.add_image
        }[ cls.dtype ]
    
    def _get_writer(cls) -> SummaryWriter():
        return GlobalSummaryWriter.getSummaryWriter()

    def __call__(cls,*args,**kwargs) -> 'Data':
        output = cls.forward(*args,**kwargs)

        # check the number or arguments required for the function. 
        # if its more than 1, pass this object as well
        for f_name,clbk in cls.callbacks.items():
            nargs = clbk.__code__.co_argcount
            if len( nargs ) > 1: cblk( output,cls,**cls.callback_kwargs[fname] )
            else: cblk( output,**cls.callback_kwargs[fname]  )
        return output
    @classmethod
    def register_callback(cls,fn : callable,**kwargs ) -> BaseInterface:
        cls.callbacks[fn.__name__] = fn
        cls.callback_kwargs[fn.__name__] = kwargs
        return cls
    @classmethod
    def unregister_callback(cls,fn : Union[callable,str]) -> BaseInterface:
        name = fn if isinstance( fn, str ) else fn.__name__
        del cls.callbacks[name]
        del cls.callback_kwargs[name]
        return cls
    @classmethod
    def forward(cls,*args,**kwargs) -> 'Data':raise NotImplementedError(f'the "__call__" method for {cls.__class__.__name__} has not been implemented!')


import torch
from metrician.interface import BaseInterface
from metrician.explainations import _predict_wrapper,get_explainer,Explainations

class BaseExplainerInterface(BaseInterface):
    """
    

    :param BaseInterface: [description]
    :type BaseInterface: [type]
    """    
    explainer_kwargs    = {}
    explainer_type      = None
    def __init__(
                    self,
                    model : torch.nn.Module
    ):
        
        # super(self.__class__,self).__init__(model)
        # self.explainer = get_explainer(self.explainer_type,self.explainer_kwargs)
        self.fn = _predict_wrapper( model )

    def _post_process(self,*args,**kwargs):raise NotImplementedError('"_post_process" not implemented for explainer!')
    def _explain(self,*args,**kwargs):raise NotImplementedError('"_explain" not implemented for explainer!')
        
    def forward(self,*args,**kwargs):
        return self._post_process(
                self._explain(*args,**kwargs)
        )




from metrician.interface import BaseInterface
BaseWriterInterface = None
class BaseWriterInterface(BaseInterface):

    montitors = {}

    def __init__(self,*args,**kwargs):
        from torch.utils.tensorboard import SummaryWriter
        self.writer = SummaryWriter()

    def register_monitor(self, kls_fn : callable ) -> BaseWriterInterface:
        self.montitors[kls_fn.__class__.__name__] = kls_fn
        return self

    def unregister_monitor(self, kls_fn : callable ) -> BaseWriterInterface:
        del self.montitors[kls_fn.__class__.__name__]
        return self

    def register_metric(self,
                        name: str, 
                        func: callable
                        ) -> 'self':
        """
            Add a callable function that will be used to evaluate the RAW data.
            If you need to coerce the data in any way, you will need to handle that on your own.

            Args:
                name: The name of the metric you want to log. This will be recorded in tensorboard
                func: The callable function that will consume the RAW data. ``i.e. (yhat,y)``
        """
        self.custom_functions[name] = func
        return self

    @staticmethod
    def from_model(model) -> BaseWriterInterface:
        from metrician.configs.interface import DefaultConfigInterface
        return BaseWriterInterface( DefaultConfigInterface.auto_config( model ) )

    def close(self):self.writer.close()

    def __exit__(self):self.close()



import torch

from .interface import BaseRecorderInterface

class Distribution(BaseRecorderInterface):
    """[summary]

    :param BaseRecorderInterface: [description]
    :type BaseRecorderInterface: [type]
    """    
    def __init__(self,reduction='mean'):
        super(self.__class__,self).__init__()
        self.reduction = reduction

    def forward(self,x:torch.Tensor) -> torch.Tensor:
        return getattr( torch,self.reduction )( x,dim=-1 )




from ..interface import BaseTestInterface

class ImageTestInterface(BaseTestInterface):

    dataset_name = ''

    _t = None

    @classmethod
    def _get_dataset(cls):
        import torchvision.datasets as dset
        import torchvision.transforms as T
        return getattr(dset,cls.dataset_name)(root='./data', train=False, transform=T.Compose([ 
                T.ToTensor(),
                # T.Resize( (32,32) )
         ]), download=True)#[:500]

    def _preprocess(self,data): return self._t( data )

    @property
    def t(self):
        assert self._t is not None, 'Image Test Transform must be defined'
        return self._t


import torch
from lime.lime_image import ImageExplanation
from .interface import BaseExplainerInterface

from collections import OrderedDict, Sequence
import numpy as np
import torch
import torch.nn as nn
from torch.nn import functional as F
from tqdm import tqdm
import matplotlib.cm as cm

class _BaseWrapper(object):
    """

    :param object: [description]
    :type object: [type]
    """    

    def __init__(self, model):
        super(_BaseWrapper, self).__init__()
        self.device = next(model.parameters()).device
        self.model = model
        self.model._forward_hooks = OrderedDict()
        self.handlers = []  # a set of hook function handlers

    def _encode_one_hot(self, ids):
        one_hot = torch.zeros_like(self.logits).to(self.device)
        one_hot.scatter_(1, ids, 1.0 )
        return one_hot

    def forward(self, image):
        """
        Simple classification
        """
        self.model.zero_grad()
        self.logits = self.model(image)
        self.probs = F.softmax(self.logits, dim=1)
        return self.probs.sort(dim=1, descending=True)

    def backward(self, ids):
        """
        Class-specific backpropagation

        Either way works:
        1. self.logits.backward(gradient=one_hot, retain_graph=True)
        2. (self.logits * one_hot).sum().backward(retain_graph=True)
        """
        one_hot = self._encode_one_hot(ids)
        self.logits.backward(gradient=one_hot, retain_graph=True)

    def generate(self):
        raise NotImplementedError

    def remove_hook(self):
        """
        Remove all the forward/backward hook functions
        """
        for handle in self.handlers:
            handle.remove()


class BackPropagation(_BaseWrapper):
    def forward(self, image):
        self.image = image.requires_grad_()
        return super(BackPropagation, self).forward(self.image)

    def generate(self):
        gradient = self.image.grad.clone()
        self.image.grad.zero_()
        return gradient


class GuidedBackPropagation(BackPropagation):
    """
    "Striving for Simplicity: the All Convolutional Net"
    https://arxiv.org/pdf/1412.6806.pdf
    Look at Figure 1 on page 8.
    """

    def __init__(self, model):
        super(GuidedBackPropagation, self).__init__(model)

        def backward_hook(module, grad_in, grad_out):
            # Cut off negative gradients
            if isinstance(module, nn.ReLU):
                return (torch.clamp(grad_in[0], min=0.0),)

        for module in self.model.named_modules():
            self.handlers.append(module[1].register_backward_hook(backward_hook))


class Deconvnet(BackPropagation):
    """
    "Striving for Simplicity: the All Convolutional Net"
    https://arxiv.org/pdf/1412.6806.pdf
    Look at Figure 1 on page 8.
    """

    def __init__(self, model):
        super(Deconvnet, self).__init__(model)

        def backward_hook(module, grad_in, grad_out):
            # Cut off negative gradients and ignore ReLU
            if isinstance(module, nn.ReLU):
                return (torch.clamp(grad_out[0], min=0.0),)

        for module in self.model.named_modules():
            self.handlers.append(module[1].register_backward_hook(backward_hook))


class GradCAM(_BaseWrapper):
    """
    "Grad-CAM: Visual Explanations from Deep Networks via Gradient-based Localization"
    https://arxiv.org/pdf/1610.02391.pdf
    Look at Figure 2 on page 4
    """

    def __init__(self, model, candidate_layers=None):
        super(GradCAM, self).__init__(model)
        self.fmap_pool = OrderedDict()
        self.grad_pool = OrderedDict()
        self.candidate_layers = candidate_layers  # list

        def forward_hook(key):
            def forward_hook_(module, input, output):
                # Save featuremaps
                self.fmap_pool[key] = output.detach()

            return forward_hook_

        def backward_hook(key):
            def backward_hook_(module, grad_in, grad_out):
                # Save the gradients correspond to the featuremaps
                self.grad_pool[key] = grad_out[0].detach()

            return backward_hook_

        # If any candidates are not specified, the hook is registered to all the layers.
        for name, module in self.model.named_modules():
            if self.candidate_layers is None or name in self.candidate_layers:
                self.handlers.append(module.register_forward_hook(forward_hook(name)))
                self.handlers.append(module.register_backward_hook(backward_hook(name)))

    def _find(self, pool, target_layer):
        if target_layer in pool.keys():
            return pool[target_layer]
        else:
            raise ValueError("Invalid layer name: {}".format(target_layer))

    def _compute_grad_weights(self, grads):
        return F.adaptive_avg_pool2d(grads, 1)

    def forward(self, image):
        
        self.image_shape = image.shape[2:]
        return super(GradCAM, self).forward(image)

    def generate(self, target_layer):
        fmaps = self._find(self.fmap_pool, target_layer)
        grads = self._find(self.grad_pool, target_layer)
        weights = self._compute_grad_weights(grads)

        gcam = torch.mul(fmaps, weights).sum(dim=1, keepdim=True)
        gcam = F.relu(gcam)

        gcam = F.interpolate(
            gcam, self.image_shape, mode="bilinear", align_corners=False
        )

        B, C, H, W = gcam.shape
        gcam = gcam.view(B, -1)
        gcam -= gcam.min(dim=1, keepdim=True)[0]
        gcam /= gcam.max(dim=1, keepdim=True)[0]
        gcam = gcam.view(B, C, H, W)

        return gcam


def occlusion_sensitivity(model, images, ids, mean=None, patch=35, stride=1, n_batches=128):
    """
    "Grad-CAM: Visual Explanations from Deep Networks via Gradient-based Localization"
    https://arxiv.org/pdf/1610.02391.pdf
    Look at Figure A5 on page 17
    
    Originally proposed in:
    "Visualizing and Understanding Convolutional Networks"
    https://arxiv.org/abs/1311.2901
    """

    torch.set_grad_enabled(False)
    model.eval()
    mean = mean if mean else 0
    patch_H, patch_W = patch if isinstance(patch, Sequence) else (patch, patch)
    pad_H, pad_W = patch_H // 2, patch_W // 2

    # Padded image
    images = F.pad(images, (pad_W, pad_W, pad_H, pad_H), value=mean)
    B, _, H, W = images.shape
    new_H = (H - patch_H) // stride + 1
    new_W = (W - patch_W) // stride + 1

    # Prepare sampling grids
    anchors = []
    grid_h = 0
    while grid_h <= H - patch_H:
        grid_w = 0
        while grid_w <= W - patch_W:
            grid_w += stride
            anchors.append((grid_h, grid_w))
        grid_h += stride

    # Baseline score without occlusion
    baseline = model(images).detach().gather(1, ids)

    # Compute per-pixel logits
    scoremaps = []
    for i in tqdm(range(0, len(anchors), n_batches), leave=False):
        batch_images = []
        batch_ids = []
        for grid_h, grid_w in anchors[i : i + n_batches]:
            images_ = images.clone()
            images_[..., grid_h : grid_h + patch_H, grid_w : grid_w + patch_W] = mean
            batch_images.append(images_)
            batch_ids.append(ids)
        batch_images = torch.cat(batch_images, dim=0)
        batch_ids = torch.cat(batch_ids, dim=0)
        scores = model(batch_images).detach().gather(1, batch_ids)
        scoremaps += list(torch.split(scores, B))

    diffmaps = torch.cat(scoremaps, dim=1) - baseline
    diffmaps = diffmaps.view(B, new_H, new_W)

    return diffmaps


def merge_grad_cam(gcam, raw_image,return_score=False, paper_cmap=False):
    from PIL import Image
    import torchvision.transforms.functional as TF
    gcam = gcam.cpu().numpy() #.squeeze(0).unsqueeze(-1)
    gcam = (cm.jet_r(gcam)[...,:3] * 255.0).squeeze(0)

    # print(raw_image);exit()
    
    v = gcam.mean(axis=1).mean(axis=0)

    l = ( (v.std()+v.mean() )/2./255.)
    # print( gcam.shape,raw_image.size() );exit()
    raw_image = raw_image.squeeze(0).permute(1,2,0).numpy()
    # if paper_cmap:
    #     alpha = gcam[..., None]
    #     gcam = alpha * cmap + ((1 - alpha) * raw_image)
    # else:
    #     # gcam = (( (gcam.astype(np.float)* l ) + ( (1.-l)* raw_image ) ) )/(raw_image+gcam).mean()
    #     # gcam = (.1*gcam)  + (.9*raw_image)
    #     # img =  raw_image*.9 + gcam
    #     gcam = Image.blend( Image.fromarray(gcam.astype(float)),Image.fromarray(raw_image.astype(float)),.5 )

    gcam = Image.blend( Image.fromarray(gcam.astype(np.uint8)), Image.fromarray((255*raw_image).astype(np.uint8)),.5)

    return gcam if not return_score else (gcam,v.std())


device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class Sparsemax(nn.Module):
    """Sparsemax function."""

    def __init__(self, dim=None):
        """Initialize sparsemax activation
        
        Args:
            dim (int, optional): The dimension over which to apply the sparsemax function.
        """
        super(Sparsemax, self).__init__()

        self.dim = -1 if dim is None else dim

    def forward(self, x):
        """Forward function.
        Args:
            x (torch.Tensor): Input tensor. First dimension should be the batch size
        Returns:
            torch.Tensor: [batch_size x number_of_logits] Output tensor
        """
        # Sparsemax currently only handles 2-dim tensors,
        # so we reshape and reshape back after sparsemax
        original_size = x.size()
        x = x.view(-1, x.size(self.dim))
        
        dim = 1
        number_of_logits = x.size(dim)

        # Translate x by max for numerical stability
        x = x - torch.max(x, dim=dim, keepdim=True)[0].expand_as(x)

        # Sort x in descending order.
        # (NOTE: Can be replaced with linear time selection method described here:
        # http://stanford.edu/~jduchi/projects/DuchiShSiCh08.html)
        zs = torch.sort(input=x, dim=dim, descending=True)[0]
        _range = torch.arange(start=1, end=number_of_logits+1, device=device).view(1, -1).type(x.type())
        _range = _range.expand_as(zs)

        # Determine sparsity of projection
        bound = 1 + _range * zs
        cumulative_sum_zs = torch.cumsum(zs, dim)
        is_gt = torch.gt(bound, cumulative_sum_zs).type(x.type())
        k = torch.max(is_gt * _range, dim, keepdim=True)[0]

        # Compute threshold function
        zs_sparse = is_gt * zs

        # Compute taus
        taus = (torch.sum(zs_sparse, dim, keepdim=True) - 1) / k
        taus = taus.expand_as(x)

        # Sparsemax
        self.output = torch.max(torch.zeros_like(x), x - taus)

        output = self.output.view(original_size)

        return output

    def backward(self, grad_output):
        """Backward function."""
        dim = 1

        nonzeros = torch.ne(self.output, 0)
        sum = torch.sum(grad_output * nonzeros, dim=dim) / torch.sum(nonzeros, dim=dim)
        self.grad_x = nonzeros * (grad_output - sum.expand_as(grad_output))

        return self.grad_x
    
    #return 

class GradCamExplainer(BaseExplainerInterface):

    def __init__(
                    self,
                    model : torch.nn.Module,
                    target_layer : str
    ):
        super(self.__class__,self).__init__(model)
        self.gcam           = GradCAM(model=model)
        self.target_layer   = target_layer

    def forward(self,image : torch.Tensor,class_label:int) -> torch.Tensor:
        """[summary]

        :param image: [description]
        :type image: torch.Tensor
        :return: [description]
        :rtype: torch.Tensor
        """        
        from pylab import figure,gca
        self.gcam.model.zero_grad()
        probs, ids = self.gcam.forward(image)
        # idxs = torch.LongTensor( [[ probs[0].argmax().item() ]] * 1 ) 
        idxs = torch.LongTensor([[ class_label ]])
        self.gcam.backward(ids=idxs)
        # self.gcam.logits.zero_grad()
        region = self.gcam.generate(self.target_layer) # region
        # region = nn.MaxPool2d(10)( region ) #torch.nn.F.grid_sample( region,  ) #transforms.ToPILImage()(region[0][0]).convert("L")
        # return region[0][0]
        img = merge_grad_cam( region[0],image )
        fig = figure()
        ax = gca()
        ax.imshow(
            img
        )
        # writer = self._get_writer()
        # writer.add_figure(
        #     'Grad Cam Explaination',
        #     fig,

        # )
        return img




class ImageExplainer(BaseExplainerInterface):
    """
    

    :param BaseExplainerInterface: [description]
    :type BaseExplainerInterface: [type]
    """    

    explainer_kwargs = {}

    def _post_process(self,explaination : ImageExplanation) -> torch.Tensor:
        pass

    def _explain(self, image : torch.Tensor ) -> ImageExplanation:
        self.explainer.explain_instance(
            image,
            num_samples=100,
            progress_bar=False
        )




import inspect
from typing import Union
from metrician.interface import BaseInterface
from metrician.signals import SignalDataKeys
class BaseSignal(BaseInterface):

    keys = [ SignalDataKeys.LOSS,SignalDataKeys.LABELS, SignalDataKeys.PREDICTIONS ]

    def _check_input(self,inputs ):
        keys = set( inputs.keys() )
        # check if keys are in the data
        if len( keys.intersection( self.keys ) ) < len(self.keys): raise AttributeError(f'Input data malformed for {self.__class__.__name__}')
        






from unittest import TestCase

class GradCamTest(TestCase):

    @classmethod
    def setUpClass(cls):
        import torchvision.models as models
        from torchvision.datasets import FakeData,CIFAR10,STL10
        # from torch.utils.data import DataLoader
        cls.model = models.resnet18(pretrained=True)
        cls.dataset = CIFAR10('./data',download=True)# FakeData() #

    def test_gradcam(self):
        from metrician.explainations.image import GradCamExplainer
        from torchvision import transforms as T
        import random
        from pylab import show
        grad_cam = GradCamExplainer(self.model,'layer3')
        grad_cam( T.ToTensor()(self.dataset[0][0]).unsqueeze(0), random.randint(0,10))


    def test_full_gradcam_pipeline(self):pass





import torch

from .interface import BaseRecorderInterface

class TopK(BaseRecorderInterface):
    """[summary]

    :param BaseRecorderInterface: [description]
    :type BaseRecorderInterface: [type]
    """    

    def __init__(self,k:int,reduction='mean'):
        super(self.__class__,self).__init__()
        self.k = k
        self.reduction = reduction


    def forward(self,x:torch.Tensor) -> torch.Tensor:
        return torch.topk( getattr( torch,self.reduction )( x,dim=-1 ),self.k)



import torch
import lime
from enum import Enum

class Explainations(Enum):
    text    = 0
    image   = 1
    tabular = 2


def get_explainer(dtype,kwargs):
    return {
        Explainations.image:lime.lime_image.LimeImageExplainer,
        # Explainations.text:lime.lime_text.LimeTextExplainer,
        # Explainations.tabular:lime.lime_tabular.LimeTabularExplainer
    }[dtype]( **kwargs )


def _predict_wrapper(func : torch.nn.Module) -> torch.nn.Module:
    '''
        provide portability to lime
    '''
    def coerce(*args,**kwargs):
        return func(*args,**kwargs).detach().cpu().numpy()
    
    return coerce



# import abc
from torch import Tensor
from enum import Enum
class ConfigType(Enum):
    unknown         =   -1
    classification  =   0
    regression      =   1
    reinforcement   =   2

class DefaultConfigInterface():#metaclass=abc.ABCMeta
    """
        Listen, I know that the class name says "Interface" and then I'm making an Abstract Class, but it's Python, give me a break.
    """
    config_type = ConfigType.unknown

    @staticmethod
    # @abc.abstractmethod
    def auto_config(model:object) -> 'DefaultConfigInterface':
        """
            Autogenerate a default config from an arbitrary model.
        """
        from metrician.src.configs.defaults import SimpleClf,SimpleRgr
        final_layer = list(model.children())[-1]
        return SimpleClf if final_layer.out_features > 1 else SimpleRgr
        

    # @abc.abstractmethod
    def sanitize_inputs(self, output : Tensor, labels : Tensor) -> (Tensor,Tensor):
        """
            For different types of models, you'll need to coerce the data differently. 
            Regression Models:
                Convert to numpy arrays and return
            Classification Models:
                Take the ArgMax (for right now), convert to a numpy array and return
        """
        return {
            ConfigType.unknown: lambda a,b : (a.clone().detach(),b.clone().detach()),#(raise ValueError('Config Type has not been set. Unable to determine how to sanitize inputs!')),
            ConfigType.classification: lambda a,b : (
                                                        a.clone().detach().argmax(dim=-1).numpy() if len( a.size() ) > 1 else a.clone().detach().numpy(),
                                                        b.clone().detach().argmax(dim=-1).numpy() if len( b.size() ) > 1 else b.clone().detach().numpy()
                                                    ),
            ConfigType.regression: lambda a,b : (a.clone().detach(),b.clone().detach())
        }[ self.config_type ]( output,labels )




    


import torch

from .interface import BaseRecorderInterface

class BertIndex2Text(BaseRecorderInterface):
    """[summary]

    :param BaseRecorderInterface: [description]
    :type BaseRecorderInterface: [type]
    """    
    def __init__(self):
        
        super(self.__class__,self).__init__()
        from transformers import PreTrainedTokenizerFast
        self.tokenizer = PreTrainedTokenizerFast()
    
    def forward(self, x : torch.Tensor ) -> str:
        """[summary]

        :param x: [description]
        :type x: torch.Tensor
        :return: [description]
        :rtype: str
        """       
        return self.tokenizer.convert_ids_to_tokens( x ) 

    